import base64
import json
import logging

from ably.sync.util.crypto import CipherData


log = logging.getLogger(__name__)


class EncodeDataMixin:

    def __init__(self, encoding):
        self.encoding = encoding

    @property
    def encoding(self):
        return '/'.join(self._encoding_array).strip('/')

    @encoding.setter
    def encoding(self, encoding):
        if not encoding:
            self._encoding_array = []
        else:
            self._encoding_array = encoding.strip('/').split('/')

    @staticmethod
    def decode(data, encoding='', cipher=None):
        encoding = encoding.strip('/')
        encoding_list = encoding.split('/')

        while encoding_list:
            encoding = encoding_list.pop()
            if not encoding:
                # With messagepack, binary data is sent as bytes, without need
                # to specify the base64 encoding. Here we coerce to bytearray,
                # since that's what is used with the Json transport; though it
                # can be argued that it should be the other way, and use always
                # bytes, never bytearray.
                if type(data) is bytes:
                    data = bytearray(data)
                continue
            if encoding == 'json':
                if isinstance(data, bytes):
                    data = data.decode()
                if isinstance(data, list) or isinstance(data, dict):
                    continue
                data = json.loads(data)
            elif encoding == 'base64' and isinstance(data, bytes):
                data = bytearray(base64.b64decode(data))
            elif encoding == 'base64':
                data = bytearray(base64.b64decode(data.encode('utf-8')))
            elif encoding.startswith('%s+' % CipherData.ENCODING_ID):
                if not cipher:
                    log.error('Message cannot be decrypted as the channel is '
                              'not set up for encryption & decryption')
                    encoding_list.append(encoding)
                    break
                data = cipher.decrypt(data)
            elif encoding == 'utf-8' and isinstance(data, (bytes, bytearray)):
                data = data.decode('utf-8')
            elif encoding == 'utf-8':
                pass
            else:
                log.error('Message cannot be decoded. '
                          "Unsupported encoding type: '%s'" % encoding)
                encoding_list.append(encoding)
                break

        encoding = '/'.join(encoding_list)
        return {'encoding': encoding, 'data': data}

    @classmethod
    def from_encoded_array(cls, objs, cipher=None):
        return [cls.from_encoded(obj, cipher=cipher) for obj in objs]



import random
import logging

from ably.transport.defaults import Defaults
from ably.types.authoptions import AuthOptions

log = logging.getLogger(__name__)


class Options(AuthOptions):
    def __init__(self, client_id=None, log_level=0, tls=True, rest_host=None, realtime_host=None, port=0,
                 tls_port=0, use_binary_protocol=True, queue_messages=False, recover=False, environment=None,
                 http_open_timeout=None, http_request_timeout=None, realtime_request_timeout=None,
                 http_max_retry_count=None, http_max_retry_duration=None, fallback_hosts=None,
                 fallback_retry_timeout=None, disconnected_retry_timeout=None, idempotent_rest_publishing=None,
                 loop=None, auto_connect=True, suspended_retry_timeout=None, connectivity_check_url=None,
                 channel_retry_timeout=Defaults.channel_retry_timeout, add_request_ids=False, **kwargs):

        super().__init__(**kwargs)

        # TODO check these defaults
        if fallback_retry_timeout is None:
            fallback_retry_timeout = Defaults.fallback_retry_timeout

        if realtime_request_timeout is None:
            realtime_request_timeout = Defaults.realtime_request_timeout

        if disconnected_retry_timeout is None:
            disconnected_retry_timeout = Defaults.disconnected_retry_timeout

        if connectivity_check_url is None:
            connectivity_check_url = Defaults.connectivity_check_url

        connection_state_ttl = Defaults.connection_state_ttl

        if suspended_retry_timeout is None:
            suspended_retry_timeout = Defaults.suspended_retry_timeout

        if environment is not None and rest_host is not None:
            raise ValueError('specify rest_host or environment, not both')

        if environment is not None and realtime_host is not None:
            raise ValueError('specify realtime_host or environment, not both')

        if idempotent_rest_publishing is None:
            from ably import api_version
            idempotent_rest_publishing = api_version >= '1.2'

        if environment is None:
            environment = Defaults.environment

        self.__client_id = client_id
        self.__log_level = log_level
        self.__tls = tls
        self.__rest_host = rest_host
        self.__realtime_host = realtime_host
        self.__port = port
        self.__tls_port = tls_port
        self.__use_binary_protocol = use_binary_protocol
        self.__queue_messages = queue_messages
        self.__recover = recover
        self.__environment = environment
        self.__http_open_timeout = http_open_timeout
        self.__http_request_timeout = http_request_timeout
        self.__realtime_request_timeout = realtime_request_timeout
        self.__http_max_retry_count = http_max_retry_count
        self.__http_max_retry_duration = http_max_retry_duration
        self.__fallback_hosts = fallback_hosts
        self.__fallback_retry_timeout = fallback_retry_timeout
        self.__disconnected_retry_timeout = disconnected_retry_timeout
        self.__channel_retry_timeout = channel_retry_timeout
        self.__idempotent_rest_publishing = idempotent_rest_publishing
        self.__loop = loop
        self.__auto_connect = auto_connect
        self.__connection_state_ttl = connection_state_ttl
        self.__suspended_retry_timeout = suspended_retry_timeout
        self.__connectivity_check_url = connectivity_check_url
        self.__fallback_realtime_host = None
        self.__add_request_ids = add_request_ids

        self.__rest_hosts = self.__get_rest_hosts()
        self.__realtime_hosts = self.__get_realtime_hosts()

    @property
    def client_id(self):
        return self.__client_id

    @client_id.setter
    def client_id(self, value):
        self.__client_id = value

    @property
    def log_level(self):
        return self.__log_level

    @log_level.setter
    def log_level(self, value):
        self.__log_level = value

    @property
    def tls(self):
        return self.__tls

    @tls.setter
    def tls(self, value):
        self.__tls = value

    @property
    def rest_host(self):
        return self.__rest_host

    @rest_host.setter
    def rest_host(self, value):
        self.__rest_host = value

    # RTC1d
    @property
    def realtime_host(self):
        return self.__realtime_host

    @realtime_host.setter
    def realtime_host(self, value):
        self.__realtime_host = value

    @property
    def port(self):
        return self.__port

    @port.setter
    def port(self, value):
        self.__port = value

    @property
    def tls_port(self):
        return self.__tls_port

    @tls_port.setter
    def tls_port(self, value):
        self.__tls_port = value

    @property
    def use_binary_protocol(self):
        return self.__use_binary_protocol

    @use_binary_protocol.setter
    def use_binary_protocol(self, value):
        self.__use_binary_protocol = value

    @property
    def queue_messages(self):
        return self.__queue_messages

    @queue_messages.setter
    def queue_messages(self, value):
        self.__queue_messages = value

    @property
    def recover(self):
        return self.__recover

    @recover.setter
    def recover(self, value):
        self.__recover = value

    @property
    def environment(self):
        return self.__environment

    @property
    def http_open_timeout(self):
        return self.__http_open_timeout

    @http_open_timeout.setter
    def http_open_timeout(self, value):
        self.__http_open_timeout = value

    @property
    def http_request_timeout(self):
        return self.__http_request_timeout

    @property
    def realtime_request_timeout(self):
        return self.__realtime_request_timeout

    @http_request_timeout.setter
    def http_request_timeout(self, value):
        self.__http_request_timeout = value

    @property
    def http_max_retry_count(self):
        return self.__http_max_retry_count

    @http_max_retry_count.setter
    def http_max_retry_count(self, value):
        self.__http_max_retry_count = value

    @property
    def http_max_retry_duration(self):
        return self.__http_max_retry_duration

    @http_max_retry_duration.setter
    def http_max_retry_duration(self, value):
        self.__http_max_retry_duration = value

    @property
    def fallback_hosts(self):
        return self.__fallback_hosts

    @property
    def fallback_retry_timeout(self):
        return self.__fallback_retry_timeout

    @property
    def disconnected_retry_timeout(self):
        return self.__disconnected_retry_timeout

    @property
    def channel_retry_timeout(self):
        return self.__channel_retry_timeout

    @property
    def idempotent_rest_publishing(self):
        return self.__idempotent_rest_publishing

    @property
    def loop(self):
        return self.__loop

    # RTC1b
    @property
    def auto_connect(self):
        return self.__auto_connect

    @property
    def connection_state_ttl(self):
        return self.__connection_state_ttl

    @connection_state_ttl.setter
    def connection_state_ttl(self, value):
        self.__connection_state_ttl = value

    @property
    def suspended_retry_timeout(self):
        return self.__suspended_retry_timeout

    @property
    def connectivity_check_url(self):
        return self.__connectivity_check_url

    @property
    def fallback_realtime_host(self):
        return self.__fallback_realtime_host

    @fallback_realtime_host.setter
    def fallback_realtime_host(self, value):
        self.__fallback_realtime_host = value

    @property
    def add_request_ids(self):
        return self.__add_request_ids

    def __get_rest_hosts(self):
        """
        Return the list of hosts as they should be tried. First comes the main
        host. Then the fallback hosts in random order.
        The returned list will have a length of up to http_max_retry_count.
        """
        # Defaults
        host = self.rest_host
        if host is None:
            host = Defaults.rest_host

        environment = self.environment

        http_max_retry_count = self.http_max_retry_count
        if http_max_retry_count is None:
            http_max_retry_count = Defaults.http_max_retry_count

        # Prepend environment
        if environment != 'production':
            host = '%s-%s' % (environment, host)

        # Fallback hosts
        fallback_hosts = self.fallback_hosts
        if fallback_hosts is None:
            if host == Defaults.rest_host:
                fallback_hosts = Defaults.fallback_hosts
            elif environment != 'production':
                fallback_hosts = Defaults.get_environment_fallback_hosts(environment)
            else:
                fallback_hosts = []

        # Shuffle
        fallback_hosts = list(fallback_hosts)
        random.shuffle(fallback_hosts)
        self.__fallback_hosts = fallback_hosts

        # First main host
        hosts = [host] + fallback_hosts
        hosts = hosts[:http_max_retry_count]
        return hosts

    def __get_realtime_hosts(self):
        if self.realtime_host is not None:
            host = self.realtime_host
            return [host]
        elif self.environment != "production":
            host = f'{self.environment}-{Defaults.realtime_host}'
        else:
            host = Defaults.realtime_host

        return [host] + self.__fallback_hosts

    def get_rest_hosts(self):
        return self.__rest_hosts

    def get_rest_host(self):
        return self.__rest_hosts[0]

    def get_realtime_hosts(self):
        return self.__realtime_hosts

    def get_realtime_host(self):
        return self.__realtime_hosts[0]

    def get_fallback_rest_hosts(self):
        return self.__rest_hosts[1:]

    def get_fallback_realtime_hosts(self):
        return self.__realtime_hosts[1:]



import asyncio
import logging
from pyee.asyncio import AsyncIOEventEmitter

from ably.sync.util.helper import is_callable_or_coroutine

# pyee's event emitter doesn't support attaching a listener to all events
# so to patch it, we create a wrapper which uses two event emitters, one
# is used to listen to all events and this arbitrary string is the event name
# used to emit all events on that listener
_all_event = 'all'

log = logging.getLogger(__name__)


def _is_named_event_args(*args):
    return len(args) == 2 and is_callable_or_coroutine(args[1])


def _is_all_event_args(*args):
    return len(args) == 1 and is_callable_or_coroutine(args[0])


class EventEmitter:
    """
    A generic interface for event registration and delivery used in a number of the types in the Realtime client
    library. For example, the Connection object emits events for connection state using the EventEmitter pattern.

    Methods
    -------
    on(*args)
        Attach to channel
    once(*args)
        Detach from channel
    off()
        Subscribe to messages on a channel
    """

    def __init__(self):
        self.__named_event_emitter = AsyncIOEventEmitter()
        self.__all_event_emitter = AsyncIOEventEmitter()
        self.__wrapped_listeners = {}

    def on(self, *args):
        """
        Registers the provided listener for the specified event, if provided, and otherwise for all events.
        If on() is called more than once with the same listener and event, the listener is added multiple times to
        its listener registry. Therefore, as an example, assuming the same listener is registered twice using
        on(), and an event is emitted once, the listener would be invoked twice.

        Parameters
        ----------
        name : str
            The named event to listen for.
        listener : callable
            The event listener.
        """
        if _is_all_event_args(*args):
            event = _all_event
            listener = args[0]
            emitter = self.__all_event_emitter
            # self.__all_event_emitter.add_listener(_all_event, args[0])
        elif _is_named_event_args(*args):
            event = args[0]
            listener = args[1]
            emitter = self.__named_event_emitter
            # self.__named_event_emitter.add_listener(args[0], args[1])
        else:
            raise ValueError("EventEmitter.on(): invalid args")

        if asyncio.iscoroutinefunction(listener):
            def wrapped_listener(*args, **kwargs):
                try:
                    listener(*args, **kwargs)
                except Exception as err:
                    log.exception(f'EventEmitter.emit(): uncaught listener exception: {err}')
        else:
            def wrapped_listener(*args, **kwargs):
                try:
                    listener(*args, **kwargs)
                except Exception as err:
                    log.exception(f'EventEmitter.emit(): uncaught listener exception: {err}')

        self.__wrapped_listeners[listener] = wrapped_listener

        emitter.add_listener(event, wrapped_listener)

    def once(self, *args):
        """
        Registers the provided listener for the first event that is emitted. If once() is called more than once
        with the same listener, the listener is added multiple times to its listener registry. Therefore, as an
        example, assuming the same listener is registered twice using once(), and an event is emitted once, the
        listener would be invoked twice. However, all subsequent events emitted would not invoke the listener as
        once() ensures that each registration is only invoked once.

        Parameters
        ----------
        name : str
            The named event to listen for.
        listener : callable
            The event listener.
        """
        if _is_all_event_args(*args):
            event = _all_event
            listener = args[0]
            emitter = self.__all_event_emitter
            # self.__all_event_emitter.add_listener(_all_event, args[0])
        elif _is_named_event_args(*args):
            event = args[0]
            listener = args[1]
            emitter = self.__named_event_emitter
            # self.__named_event_emitter.add_listener(args[0], args[1])
        else:
            raise ValueError("EventEmitter.on(): invalid args")

        if asyncio.iscoroutinefunction(listener):
            def wrapped_listener(*args, **kwargs):
                try:
                    listener(*args, **kwargs)
                except Exception as err:
                    log.exception(f'EventEmitter.emit(): uncaught listener exception: {err}')
        else:
            def wrapped_listener(*args, **kwargs):
                try:
                    listener(*args, **kwargs)
                except Exception as err:
                    log.exception(f'EventEmitter.emit(): uncaught listener exception: {err}')

        self.__wrapped_listeners[listener] = wrapped_listener

        emitter.once(event, wrapped_listener)

    def off(self, *args):
        """
        Removes all registrations that match both the specified listener and, if provided, the specified event.
        If called with no arguments, deregisters all registrations, for all events and listeners.

        Parameters
        ----------
        name : str
            The named event to listen for.
        listener : callable
            The event listener.
        """
        if len(args) == 0:
            self.__all_event_emitter.remove_all_listeners()
            self.__named_event_emitter.remove_all_listeners()
            return
        elif _is_all_event_args(*args):
            event = _all_event
            listener = args[0]
            emitter = self.__all_event_emitter
        elif _is_named_event_args(*args):
            event = args[0]
            listener = args[1]
            emitter = self.__named_event_emitter
        else:
            raise ValueError("EventEmitter.once(): invalid args")

        wrapped_listener = self.__wrapped_listeners.get(listener)

        if wrapped_listener is None:
            return

        emitter.remove_listener(event, wrapped_listener)
        self.__wrapped_listeners[listener] = None

    def once_async(self, state=None):
        future = asyncio.Future()

        def on_state_change(*args):
            future.set_result(*args)

        if state is not None:
            self.once(state, on_state_change)
        else:
            self.once(on_state_change)

        state_change = future

        return state_change

    def _emit(self, *args):
        self.__named_event_emitter.emit(*args)
        self.__all_event_emitter.emit(_all_event, *args[1:])



from __future__ import annotations
from typing import TYPE_CHECKING
import asyncio
from enum import IntEnum
import json
import logging
import socket
import urllib.parse
from ably.sync.http.httputils import HttpUtils
from ably.sync.types.connectiondetails import ConnectionDetails
from ably.sync.util.eventemitter import EventEmitter
from ably.sync.util.exceptions import AblyException
from ably.sync.util.helper import Timer, unix_time_ms
from websockets.client import WebSocketClientProtocol, connect as ws_connect
from websockets.exceptions import ConnectionClosedOK, WebSocketException

if TYPE_CHECKING:
    from ably.sync.realtime.connection import ConnectionManager

log = logging.getLogger(__name__)


class ProtocolMessageAction(IntEnum):
    HEARTBEAT = 0
    CONNECTED = 4
    DISCONNECTED = 6
    CLOSE = 7
    CLOSED = 8
    ERROR = 9
    ATTACH = 10
    ATTACHED = 11
    DETACH = 12
    DETACHED = 13
    MESSAGE = 15
    AUTH = 17


class WebSocketTransport(EventEmitter):
    def __init__(self, connection_manager: ConnectionManager, host: str, params: dict):
        self.websocket: WebSocketClientProtocol | None = None
        self.read_loop: asyncio.Task | None = None
        self.connect_task: asyncio.Task | None = None
        self.ws_connect_task: asyncio.Task | None = None
        self.connection_manager = connection_manager
        self.options = self.connection_manager.options
        self.is_connected = False
        self.idle_timer = None
        self.last_activity = None
        self.max_idle_interval = None
        self.is_disposed = False
        self.host = host
        self.params = params
        super().__init__()

    def connect(self):
        headers = HttpUtils.default_headers()
        query_params = urllib.parse.urlencode(self.params)
        ws_url = (f'wss://{self.host}?{query_params}')
        log.info(f'connect(): attempting to connect to {ws_url}')
        self.ws_connect_task = asyncio.create_task(self.ws_connect(ws_url, headers))
        self.ws_connect_task.add_done_callback(self.on_ws_connect_done)

    def on_ws_connect_done(self, task: asyncio.Task):
        try:
            exception = task.exception()
        except asyncio.CancelledError as e:
            exception = e
        if exception is None or isinstance(exception, ConnectionClosedOK):
            return
        log.info(
            f'WebSocketTransport.on_ws_connect_done(): exception = {exception}'
        )

    def ws_connect(self, ws_url, headers):
        try:
            with ws_connect(ws_url, extra_headers=headers) as websocket:
                log.info(f'ws_connect(): connection established to {ws_url}')
                self._emit('connected')
                self.websocket = websocket
                self.read_loop = self.connection_manager.options.loop.create_task(self.ws_read_loop())
                self.read_loop.add_done_callback(self.on_read_loop_done)
                try:
                    self.read_loop
                except WebSocketException as err:
                    if not self.is_disposed:
                        self.dispose()
                        self.connection_manager.deactivate_transport(err)
        except (WebSocketException, socket.gaierror) as e:
            exception = AblyException(f'Error opening websocket connection: {e}', 400, 40000)
            log.exception(f'WebSocketTransport.ws_connect(): Error opening websocket connection: {exception}')
            self._emit('failed', exception)
            raise exception

    def on_protocol_message(self, msg):
        self.on_activity()
        log.debug(f'WebSocketTransport.on_protocol_message(): received protocol message: {msg}')
        action = msg.get('action')
        if action == ProtocolMessageAction.CONNECTED:
            connection_id = msg.get('connectionId')
            connection_details = ConnectionDetails.from_dict(msg.get('connectionDetails'))

            error = msg.get('error')
            exception = None
            if error:
                exception = AblyException.from_dict(error)

            max_idle_interval = connection_details.max_idle_interval
            if max_idle_interval:
                self.max_idle_interval = max_idle_interval + self.options.realtime_request_timeout
                self.on_activity()
            self.is_connected = True
            if self.host != self.options.get_realtime_host():  # RTN17e
                self.options.fallback_realtime_host = self.host
            self.connection_manager.on_connected(connection_details, connection_id, reason=exception)
        elif action == ProtocolMessageAction.DISCONNECTED:
            error = msg.get('error')
            exception = None
            if error is not None:
                exception = AblyException.from_dict(error)
            self.connection_manager.on_disconnected(exception)
        elif action == ProtocolMessageAction.AUTH:
            try:
                self.connection_manager.ably.auth.authorize()
            except Exception as exc:
                log.exception(f"WebSocketTransport.on_protocol_message(): An exception \
                                occurred during reauth: {exc}")
        elif action == ProtocolMessageAction.CLOSED:
            if self.ws_connect_task:
                self.ws_connect_task.cancel()
            self.connection_manager.on_closed()
        elif action == ProtocolMessageAction.ERROR:
            error = msg.get('error')
            exception = AblyException.from_dict(error)
            self.connection_manager.on_error(msg, exception)
        elif action == ProtocolMessageAction.HEARTBEAT:
            id = msg.get('id')
            self.connection_manager.on_heartbeat(id)
        elif action in (
            ProtocolMessageAction.ATTACHED,
            ProtocolMessageAction.DETACHED,
            ProtocolMessageAction.MESSAGE
        ):
            self.connection_manager.on_channel_message(msg)

    def ws_read_loop(self):
        if not self.websocket:
            raise AblyException('ws_read_loop started with no websocket', 500, 50000)
        try:
            for raw in self.websocket:
                msg = json.loads(raw)
                task = asyncio.create_task(self.on_protocol_message(msg))
                task.add_done_callback(self.on_protcol_message_handled)
        except ConnectionClosedOK:
            return

    def on_protcol_message_handled(self, task):
        try:
            exception = task.exception()
        except Exception as e:
            exception = e
        if exception is not None:
            log.exception(f"WebSocketTransport.on_protocol_message_handled(): uncaught exception: {exception}")

    def on_read_loop_done(self, task: asyncio.Task):
        try:
            exception = task.exception()
        except asyncio.CancelledError as e:
            exception = e
        if isinstance(exception, ConnectionClosedOK):
            return

    def dispose(self):
        self.is_disposed = True
        if self.read_loop:
            self.read_loop.cancel()
        if self.ws_connect_task:
            self.ws_connect_task.cancel()
        if self.idle_timer:
            self.idle_timer.cancel()
        if self.websocket:
            try:
                self.websocket.close()
            except asyncio.CancelledError:
                return

    def close(self):
        self.send({'action': ProtocolMessageAction.CLOSE})

    def send(self, message: dict):
        if self.websocket is None:
            raise Exception()
        raw_msg = json.dumps(message)
        log.info(f'WebSocketTransport.send(): sending {raw_msg}')
        self.websocket.send(raw_msg)

    def set_idle_timer(self, timeout: float):
        if not self.idle_timer:
            self.idle_timer = Timer(timeout, self.on_idle_timer_expire)

    def on_idle_timer_expire(self):
        self.idle_timer = None
        since_last = unix_time_ms() - self.last_activity
        time_remaining = self.max_idle_interval - since_last
        msg = f"No activity seen from realtime in {since_last} ms; assuming connection has dropped"
        if time_remaining <= 0:
            log.error(msg)
            self.disconnect(AblyException(msg, 408, 80003))
        else:
            self.set_idle_timer(time_remaining + 100)

    def on_activity(self):
        if not self.max_idle_interval:
            return
        self.last_activity = unix_time_ms()
        self.set_idle_timer(self.max_idle_interval + 100)

    def disconnect(self, reason=None):
        self.dispose()
        self.connection_manager.deactivate_transport(reason)



import glob
import os
import tokenize as std_tokenize

import tokenize_rt

rename_classes = [
    "AblyRest",
    "Push",
    "PushAdmin",
    "Channel",
    "Channels",
    "AuthSync",
    "Http",
    "PaginatedResult",
    "HttpPaginatedResponse"
]

_TOKEN_REPLACE = {
    "__aenter__": "__enter__",
    "__aexit__": "__exit__",
    "__aiter__": "__iter__",
    "__anext__": "__next__",
    "asynccontextmanager": "contextmanager",
    "AsyncIterable": "Iterable",
    "AsyncIterator": "Iterator",
    "AsyncGenerator": "Generator",
    "StopAsyncIteration": "StopIteration",
}

_IMPORTS_REPLACE = {
}

_STRING_REPLACE = {
}

_CLASS_RENAME = {
}


class Rule:
    """A single set of rules for 'unasync'ing file(s)"""

    def __init__(self, fromdir, todir, output_file_prefix="", additional_replacements=None):
        self.fromdir = fromdir.replace("/", os.sep)
        self.todir = todir.replace("/", os.sep)
        self.ouput_file_prefix = output_file_prefix

        # Add any additional user-defined token replacements to our list.
        self.token_replacements = _TOKEN_REPLACE.copy()
        for key, val in (additional_replacements or {}).items():
            self.token_replacements[key] = val

    def _match(self, filepath):
        """Determines if a Rule matches a given filepath and if so
        returns a higher comparable value if the match is more specific.
        """
        file_segments = [x for x in filepath.split(os.sep) if x]
        from_segments = [x for x in self.fromdir.split(os.sep) if x]
        len_from_segments = len(from_segments)

        if len_from_segments > len(file_segments):
            return False

        for i in range(len(file_segments) - len_from_segments + 1):
            if file_segments[i: i + len_from_segments] == from_segments:
                return len_from_segments, i

        return False

    def _unasync_file(self, filepath):
        with open(filepath, "rb") as f:
            encoding, _ = std_tokenize.detect_encoding(f.readline)

        with open(filepath, "rt", encoding=encoding) as f:
            tokens = tokenize_rt.src_to_tokens(f.read())
            tokens = self._unasync_tokens(tokens)
            result = tokenize_rt.tokens_to_src(tokens)
            new_file_path = os.path.join(os.path.dirname(filepath),
                                         self.ouput_file_prefix + os.path.basename(filepath))
            outfilepath = new_file_path.replace(self.fromdir, self.todir)
            os.makedirs(os.path.dirname(outfilepath), exist_ok=True)
            with open(outfilepath, "wb") as f:
                f.write(result.encode(encoding))

    def _unasync_tokens(self, tokens: list):
        new_tokens = []
        token_counter = 0
        async_await_block_started = False
        async_await_char_diff = -6  # (len("async") or len("await") is 6)
        async_await_offset = 0

        renamed_class_call_started = False
        renamed_class_char_diff = 0
        renamed_class_offset = 0

        while token_counter < len(tokens):
            token = tokens[token_counter]

            if async_await_block_started or renamed_class_call_started:
                # Fix indentation issues for async/await fn definition/call
                if token.src == '\n':
                    new_tokens.append(token)
                    token_counter = token_counter + 1
                    next_newline_token = tokens[token_counter]
                    new_tab_src = next_newline_token.src

                    if (renamed_class_call_started and
                            tokens[token_counter + 1].utf8_byte_offset >= renamed_class_offset):
                        if renamed_class_char_diff < 0:
                            new_tab_src = new_tab_src[:renamed_class_char_diff]
                        else:
                            new_tab_src = new_tab_src + renamed_class_char_diff * " "

                    if (async_await_block_started and len(next_newline_token.src) >= 6 and
                            tokens[token_counter + 1].utf8_byte_offset >= async_await_offset + 6):
                        new_tab_src = new_tab_src[:async_await_char_diff]  # remove last 6 white spaces

                    next_newline_token = next_newline_token._replace(src=new_tab_src)
                    new_tokens.append(next_newline_token)
                    token_counter = token_counter + 1
                    continue

            if token.src == ')':
                async_await_block_started = False
                async_await_offset = 0
                renamed_class_call_started = False
                renamed_class_char_diff = 0

            if token.src in ["async", "await"]:
                # When removing async or await, we want to skip the following whitespace
                token_counter = token_counter + 2
                is_async_start = tokens[token_counter].src == 'def'
                is_await_start = False
                for i in range(token_counter, token_counter + 6):
                    if tokens[i].src == '(':
                        is_await_start = True
                        break
                if is_async_start or is_await_start:
                    # Fix indentation issues for async/await fn definition/call
                    async_await_offset = token.utf8_byte_offset
                    async_await_block_started = True
                continue

            elif token.name == "NAME":
                if token.src == "from":
                    if tokens[token_counter + 1].src == " ":
                        token_counter = self._replace_import(tokens, token_counter, new_tokens)
                        continue
                else:
                    token_new_src = self._unasync_name(token.src)
                    if token.src == token_new_src:
                        token_new_src = self._class_rename(token.src)
                        if token.src != token_new_src:
                            renamed_class_offset = token.utf8_byte_offset
                            renamed_class_char_diff = len(token_new_src) - len(token.src)
                            for i in range(token_counter, token_counter + 6):
                                if tokens[i].src == '(':
                                    renamed_class_call_started = True
                                    break

                    token = token._replace(src=token_new_src)
            elif token.name == "STRING":
                src_token = token.src.replace("'", "")
                if _STRING_REPLACE.get(src_token) is not None:
                    new_token = f"'{_STRING_REPLACE[src_token]}'"
                    token = token._replace(src=new_token)
                else:
                    src_token = token.src.replace("\"", "")
                    if _STRING_REPLACE.get(src_token) is not None:
                        new_token = f"\"{_STRING_REPLACE[src_token]}\""
                        token = token._replace(src=new_token)

            new_tokens.append(token)
            token_counter = token_counter + 1

        return new_tokens

    def _replace_import(self, tokens, token_counter, new_tokens: list):
        new_tokens.append(tokens[token_counter])
        new_tokens.append(tokens[token_counter + 1])

        full_lib_name = ''
        lib_name_counter = token_counter + 2
        if len(_IMPORTS_REPLACE.keys()) == 0:
            return lib_name_counter

        while True:
            if tokens[lib_name_counter].src == " ":
                break
            full_lib_name = full_lib_name + tokens[lib_name_counter].src
            lib_name_counter = lib_name_counter + 1

        for key, value in _IMPORTS_REPLACE.items():
            if key in full_lib_name:
                updated_lib_name = full_lib_name.replace(key, value)
                for lib_name_part in updated_lib_name.split("."):
                    lib_name_part = self._class_rename(lib_name_part)
                    new_tokens.append(tokenize_rt.Token("NAME", lib_name_part))
                    new_tokens.append(tokenize_rt.Token("OP", "."))
                new_tokens.pop()
                return lib_name_counter

        lib_name_counter = token_counter + 2
        return lib_name_counter

    def _class_rename(self, name):
        if name in _CLASS_RENAME:
            return _CLASS_RENAME[name]
        return name

    def _unasync_name(self, name):
        if name in self.token_replacements:
            return self.token_replacements[name]
        return name


def unasync_files(fpath_list, rules):
    for f in fpath_list:
        found_rule = None
        found_weight = None

        for rule in rules:
            weight = rule._match(f)
            if weight and (found_weight is None or weight > found_weight):
                found_rule = rule
                found_weight = weight

        if found_rule:
            found_rule._unasync_file(f)


def find_files(dir_path, file_name_regex):
    return glob.glob(os.path.join(dir_path, "**", file_name_regex), recursive=True)


def run():
    # Source files ==========================================

    _TOKEN_REPLACE["AsyncClient"] = "Client"
    _TOKEN_REPLACE["aclose"] = "close"

    _IMPORTS_REPLACE["ably"] = "ably.sync"

    # here...
    for class_name in rename_classes:
        _CLASS_RENAME[class_name] = f"{class_name}Sync"

    _STRING_REPLACE["AuthSync"] = "AuthSync"

    src_dir_path = os.path.join(os.getcwd(), "ably")
    dest_dir_path = os.path.join(os.getcwd(), "ably", "sync")

    relevant_src_files = (set(find_files(src_dir_path, "*.py")) -
                          set(find_files(dest_dir_path, "*.py")))

    unasync_files(list(relevant_src_files), [Rule(fromdir=src_dir_path, todir=dest_dir_path)])

    # Test files ==============================================

    _TOKEN_REPLACE["asyncSetUp"] = "setUp"
    _TOKEN_REPLACE["asyncTearDown"] = "tearDown"
    _TOKEN_REPLACE["AsyncMock"] = "Mock"

    _TOKEN_REPLACE["_Channel__publish_request_body"] = "_ChannelSync__publish_request_body"
    _TOKEN_REPLACE["_Http__client"] = "_HttpSync__client"

    _IMPORTS_REPLACE["test.ably"] = "test.ably.sync"

    _STRING_REPLACE['/../assets/testAppSpec.json'] = '/../../assets/testAppSpec.json'
    _STRING_REPLACE['ably.rest.auth.Auth.request_token'] = 'ably.sync.rest.auth.AuthSync.request_token'
    _STRING_REPLACE['ably.rest.auth.TokenRequest'] = 'ably.sync.rest.auth.TokenRequest'
    _STRING_REPLACE['ably.rest.rest.Http.post'] = 'ably.sync.rest.rest.HttpSync.post'
    _STRING_REPLACE['httpx.AsyncClient.send'] = 'httpx.Client.send'
    _STRING_REPLACE['ably.util.exceptions.AblyException.raise_for_response'] = \
        'ably.sync.util.exceptions.AblyException.raise_for_response'
    _STRING_REPLACE['ably.rest.rest.AblyRest.time'] = 'ably.sync.rest.rest.AblyRestSync.time'
    _STRING_REPLACE['ably.rest.auth.Auth._timestamp'] = 'ably.sync.rest.auth.AuthSync._timestamp'

    # round 1
    src_dir_path = os.path.join(os.getcwd(), "test", "ably")
    dest_dir_path = os.path.join(os.getcwd(), "test", "ably", "sync")
    src_files = [os.path.join(os.getcwd(), "test", "ably", "testapp.py"),
                 os.path.join(os.getcwd(), "test", "ably", "utils.py")]

    unasync_files(src_files, [Rule(fromdir=src_dir_path, todir=dest_dir_path)])

    # round 2
    src_dir_path = os.path.join(os.getcwd(), "test", "ably", "rest")
    dest_dir_path = os.path.join(os.getcwd(), "test", "ably", "sync", "rest")
    src_files = find_files(src_dir_path, "*.py")

    unasync_files(src_files, [Rule(fromdir=src_dir_path, todir=dest_dir_path, output_file_prefix="sync_")])



from ably.rest.rest import AblyRest
from ably.realtime.realtime import AblyRealtime
from ably.rest.auth import Auth
from ably.rest.push import Push
from ably.types.capability import Capability
from ably.types.channelsubscription import PushChannelSubscription
from ably.types.device import DeviceDetails
from ably.types.options import Options
from ably.util.crypto import CipherParams
from ably.util.exceptions import AblyException, AblyAuthException, IncompatibleClientIdException

import logging

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

api_version = '3'
lib_version = '2.0.2'



from typing import Optional
from ably.http.paginatedresult import PaginatedResult, format_params
from ably.types.device import DeviceDetails, device_details_response_processor
from ably.types.channelsubscription import PushChannelSubscription, channel_subscriptions_response_processor
from ably.types.channelsubscription import channels_response_processor


class Push:

    def __init__(self, ably):
        self.__ably = ably
        self.__admin = PushAdmin(ably)

    @property
    def admin(self):
        return self.__admin


class PushAdmin:

    def __init__(self, ably):
        self.__ably = ably
        self.__device_registrations = PushDeviceRegistrations(ably)
        self.__channel_subscriptions = PushChannelSubscriptions(ably)

    @property
    def ably(self):
        return self.__ably

    @property
    def device_registrations(self):
        return self.__device_registrations

    @property
    def channel_subscriptions(self):
        return self.__channel_subscriptions

    async def publish(self, recipient: dict, data: dict, timeout: Optional[float] = None):
        """Publish a push notification to a single device.

        :Parameters:
        - `recipient`: the recipient of the notification
        - `data`: the data of the notification
        """
        if not isinstance(recipient, dict):
            raise TypeError('Unexpected %s recipient, expected a dict' % type(recipient))

        if not isinstance(data, dict):
            raise TypeError('Unexpected %s data, expected a dict' % type(data))

        if not recipient:
            raise ValueError('recipient is empty')

        if not data:
            raise ValueError('data is empty')

        body = data.copy()
        body.update({'recipient': recipient})
        await self.ably.http.post('/push/publish', body=body, timeout=timeout)


class PushDeviceRegistrations:

    def __init__(self, ably):
        self.__ably = ably

    @property
    def ably(self):
        return self.__ably

    async def get(self, device_id: str):
        """Returns a DeviceDetails object if the device id is found or results
        in a not found error if the device cannot be found.

        :Parameters:
        - `device_id`: the id of the device
        """
        path = '/push/deviceRegistrations/%s' % device_id
        response = await self.ably.http.get(path)
        obj = response.to_native()
        return DeviceDetails.from_dict(obj)

    async def list(self, **params):
        """Returns a PaginatedResult object with the list of DeviceDetails
        objects, filtered by the given parameters.

        :Parameters:
        - `**params`: the parameters used to filter the list
        """
        path = '/push/deviceRegistrations' + format_params(params)
        return await PaginatedResult.paginated_query(
            self.ably.http, url=path,
            response_processor=device_details_response_processor)

    async def save(self, device: dict):
        """Creates or updates the device. Returns a DeviceDetails object.

        :Parameters:
        - `device`: a dictionary with the device information
        """
        device_details = DeviceDetails.factory(device)
        path = '/push/deviceRegistrations/%s' % device_details.id
        body = device_details.as_dict()
        response = await self.ably.http.put(path, body=body)
        obj = response.to_native()
        return DeviceDetails.from_dict(obj)

    async def remove(self, device_id: str):
        """Deletes the registered device identified by the given device id.

        :Parameters:
        - `device_id`: the id of the device
        """
        path = '/push/deviceRegistrations/%s' % device_id
        return await self.ably.http.delete(path)

    async def remove_where(self, **params):
        """Deletes the registered devices identified by the given parameters.

        :Parameters:
        - `**params`: the parameters that identify the devices to remove
        """
        path = '/push/deviceRegistrations' + format_params(params)
        return await self.ably.http.delete(path)


class PushChannelSubscriptions:

    def __init__(self, ably):
        self.__ably = ably

    @property
    def ably(self):
        return self.__ably

    async def list(self, **params):
        """Returns a PaginatedResult object with the list of
        PushChannelSubscription objects, filtered by the given parameters.

        :Parameters:
        - `**params`: the parameters used to filter the list
        """
        path = '/push/channelSubscriptions' + format_params(params)
        return await PaginatedResult.paginated_query(self.ably.http, url=path,
                                                     response_processor=channel_subscriptions_response_processor)

    async def list_channels(self, **params):
        """Returns a PaginatedResult object with the list of
        PushChannelSubscription objects, filtered by the given parameters.

        :Parameters:
        - `**params`: the parameters used to filter the list
        """
        path = '/push/channels' + format_params(params)
        return await PaginatedResult.paginated_query(self.ably.http, url=path,
                                                     response_processor=channels_response_processor)

    async def save(self, subscription: dict):
        """Creates or updates the subscription. Returns a
        PushChannelSubscription object.

        :Parameters:
        - `subscription`: a dictionary with the subscription information
        """
        subscription = PushChannelSubscription.factory(subscription)
        path = '/push/channelSubscriptions'
        body = subscription.as_dict()
        response = await self.ably.http.post(path, body=body)
        obj = response.to_native()
        return PushChannelSubscription.from_dict(obj)

    async def remove(self, subscription: dict):
        """Deletes the given subscription.

        :Parameters:
        - `subscription`: the subscription object to remove
        """
        subscription = PushChannelSubscription.factory(subscription)
        params = subscription.as_dict()
        return await self.remove_where(**params)

    async def remove_where(self, **params):
        """Deletes the subscriptions identified by the given parameters.

        :Parameters:
        - `**params`: the parameters that identify the subscriptions to remove
        """
        path = '/push/channelSubscriptions' + format_params(**params)
        return await self.ably.http.delete(path)



from ably.util.exceptions import AblyException


class AuthOptions:
    def __init__(self, auth_callback=None, auth_url=None, auth_method='GET',
                 auth_token=None, auth_headers=None, auth_params=None,
                 key_name=None, key_secret=None, key=None, query_time=False,
                 token_details=None, use_token_auth=None,
                 default_token_params=None):
        self.__auth_options = {}
        self.auth_options['auth_callback'] = auth_callback
        self.auth_options['auth_url'] = auth_url
        self.auth_options['auth_method'] = auth_method
        self.auth_options['auth_headers'] = auth_headers
        self.auth_options['auth_params'] = auth_params
        self.auth_options['query_time'] = query_time
        self.auth_options['key_name'] = key_name
        self.auth_options['key_secret'] = key_secret
        self.set_key(key)

        self.__auth_token = auth_token
        self.__token_details = token_details
        self.__use_token_auth = use_token_auth
        default_token_params = default_token_params or {}
        default_token_params.pop('timestamp', None)
        self.default_token_params = default_token_params

    def set_key(self, key):
        if key is None:
            return

        try:
            key_name, key_secret = key.split(':')
            self.auth_options['key_name'] = key_name
            self.auth_options['key_secret'] = key_secret
        except ValueError:
            raise AblyException("key of not len 2 parameters: {0}"
                                .format(key.split(':')),
                                401, 40101)

    def replace(self, auth_options):
        if type(auth_options) is dict:
            auth_options = dict(auth_options)
            key = auth_options.pop('key', None)
            self.auth_options = auth_options
            self.set_key(key)
        elif type(auth_options) is AuthOptions:
            self.auth_options = dict(auth_options.auth_options)
        else:
            raise KeyError('Expected dict or AuthOptions')

    @property
    def auth_options(self):
        return self.__auth_options

    @auth_options.setter
    def auth_options(self, value):
        self.__auth_options = value

    @property
    def auth_callback(self):
        return self.auth_options['auth_callback']

    @auth_callback.setter
    def auth_callback(self, value):
        self.auth_options['auth_callback'] = value

    @property
    def auth_url(self):
        return self.auth_options['auth_url']

    @auth_url.setter
    def auth_url(self, value):
        self.auth_options['auth_url'] = value

    @property
    def auth_method(self):
        return self.auth_options['auth_method']

    @auth_method.setter
    def auth_method(self, value):
        self.auth_options['auth_method'] = value.upper()

    @property
    def key_name(self):
        return self.auth_options['key_name']

    @key_name.setter
    def key_name(self, value):
        self.auth_options['key_name'] = value

    @property
    def key_secret(self):
        return self.auth_options['key_secret']

    @key_secret.setter
    def key_secret(self, value):
        self.auth_options['key_secret'] = value

    @property
    def auth_token(self):
        return self.__auth_token

    @auth_token.setter
    def auth_token(self, value):
        self.__auth_token = value

    @property
    def auth_headers(self):
        return self.auth_options['auth_headers']

    @auth_headers.setter
    def auth_headers(self, value):
        self.auth_options['auth_headers'] = value

    @property
    def auth_params(self):
        return self.auth_options['auth_params']

    @auth_params.setter
    def auth_params(self, value):
        self.auth_options['auth_params'] = value

    @property
    def query_time(self):
        return self.auth_options['query_time']

    @query_time.setter
    def query_time(self, value):
        self.auth_options['query_time'] = value

    @property
    def token_details(self):
        return self.__token_details

    @token_details.setter
    def token_details(self, value):
        self.__token_details = value

    @property
    def use_token_auth(self):
        return self.__use_token_auth

    @use_token_auth.setter
    def use_token_auth(self, value):
        self.__use_token_auth = value

    @property
    def default_token_params(self):
        return self.__default_token_params

    @default_token_params.setter
    def default_token_params(self, value):
        self.__default_token_params = value

    def __str__(self):
        return str(self.__dict__)



from __future__ import annotations
import logging
import asyncio
import httpx
from ably.transport.websockettransport import WebSocketTransport, ProtocolMessageAction
from ably.transport.defaults import Defaults
from ably.types.connectionerrors import ConnectionErrors
from ably.types.connectionstate import ConnectionEvent, ConnectionState, ConnectionStateChange
from ably.types.tokendetails import TokenDetails
from ably.util.exceptions import AblyException, IncompatibleClientIdException
from ably.util.eventemitter import EventEmitter
from datetime import datetime
from ably.util.helper import get_random_id, Timer, is_token_error
from typing import Optional, TYPE_CHECKING
from ably.types.connectiondetails import ConnectionDetails
from queue import Queue

if TYPE_CHECKING:
    from ably.realtime.realtime import AblyRealtime

log = logging.getLogger(__name__)


class ConnectionManager(EventEmitter):
    def __init__(self, realtime: AblyRealtime, initial_state):
        self.options = realtime.options
        self.__ably = realtime
        self.__state: ConnectionState = initial_state
        self.__ping_future: Optional[asyncio.Future] = None
        self.__timeout_in_secs: float = self.options.realtime_request_timeout / 1000
        self.transport: Optional[WebSocketTransport] = None
        self.__connection_details: Optional[ConnectionDetails] = None
        self.connection_id: Optional[str] = None
        self.__fail_state = ConnectionState.DISCONNECTED
        self.transition_timer: Optional[Timer] = None
        self.suspend_timer: Optional[Timer] = None
        self.retry_timer: Optional[Timer] = None
        self.connect_base_task: Optional[asyncio.Task] = None
        self.disconnect_transport_task: Optional[asyncio.Task] = None
        self.__fallback_hosts: list[str] = self.options.get_fallback_realtime_hosts()
        self.queued_messages: Queue = Queue()
        self.__error_reason: Optional[AblyException] = None
        super().__init__()

    def enact_state_change(self, state: ConnectionState, reason: Optional[AblyException] = None) -> None:
        current_state = self.__state
        log.debug(f'ConnectionManager.enact_state_change(): {current_state} -> {state}; reason = {reason}')
        self.__state = state
        if reason:
            self.__error_reason = reason
        self._emit('connectionstate', ConnectionStateChange(current_state, state, state, reason))

    def check_connection(self) -> bool:
        try:
            response = httpx.get(self.options.connectivity_check_url)
            return 200 <= response.status_code < 300 and \
                (self.options.connectivity_check_url != Defaults.connectivity_check_url or "yes" in response.text)
        except httpx.HTTPError:
            return False

    def get_state_error(self) -> AblyException:
        return ConnectionErrors[self.state]

    async def __get_transport_params(self) -> dict:
        protocol_version = Defaults.protocol_version
        params = await self.ably.auth.get_auth_transport_param()
        params["v"] = protocol_version
        if self.connection_details:
            params["resume"] = self.connection_details.connection_key
        return params

    async def close_impl(self) -> None:
        log.debug('ConnectionManager.close_impl()')

        self.cancel_suspend_timer()
        self.start_transition_timer(ConnectionState.CLOSING, fail_state=ConnectionState.CLOSED)
        if self.transport:
            await self.transport.dispose()
        if self.connect_base_task:
            self.connect_base_task.cancel()
        if self.disconnect_transport_task:
            await self.disconnect_transport_task
        self.cancel_retry_timer()

        self.notify_state(ConnectionState.CLOSED)

    async def send_protocol_message(self, protocol_message: dict) -> None:
        if self.state in (
            ConnectionState.DISCONNECTED,
            ConnectionState.CONNECTING,
        ):
            self.queued_messages.put(protocol_message)
            return

        if self.state == ConnectionState.CONNECTED:
            if self.transport:
                await self.transport.send(protocol_message)
            else:
                log.exception(
                    "ConnectionManager.send_protocol_message(): can not send message with no active transport"
                )
            return

        raise AblyException(f"ConnectionManager.send_protocol_message(): called in {self.state}", 500, 50000)

    def send_queued_messages(self) -> None:
        log.info(f'ConnectionManager.send_queued_messages(): sending {self.queued_messages.qsize()} message(s)')
        while not self.queued_messages.empty():
            asyncio.create_task(self.send_protocol_message(self.queued_messages.get()))

    def fail_queued_messages(self, err) -> None:
        log.info(
            f"ConnectionManager.fail_queued_messages(): discarding {self.queued_messages.qsize()} messages;" +
            f" reason = {err}"
        )
        while not self.queued_messages.empty():
            msg = self.queued_messages.get()
            log.exception(f"ConnectionManager.fail_queued_messages(): Failed to send protocol message: {msg}")

    async def ping(self) -> float:
        if self.__ping_future:
            try:
                response = await self.__ping_future
            except asyncio.CancelledError:
                raise AblyException("Ping request cancelled due to request timeout", 504, 50003)
            return response

        self.__ping_future = asyncio.Future()
        if self.__state in [ConnectionState.CONNECTED, ConnectionState.CONNECTING]:
            self.__ping_id = get_random_id()
            ping_start_time = datetime.now().timestamp()
            await self.send_protocol_message({"action": ProtocolMessageAction.HEARTBEAT,
                                              "id": self.__ping_id})
        else:
            raise AblyException("Cannot send ping request. Calling ping in invalid state", 40000, 400)
        try:
            await asyncio.wait_for(self.__ping_future, self.__timeout_in_secs)
        except asyncio.TimeoutError:
            raise AblyException("Timeout waiting for ping response", 504, 50003)

        ping_end_time = datetime.now().timestamp()
        response_time_ms = (ping_end_time - ping_start_time) * 1000
        return round(response_time_ms, 2)

    def on_connected(self, connection_details: ConnectionDetails, connection_id: str,
                     reason: Optional[AblyException] = None) -> None:
        self.__fail_state = ConnectionState.DISCONNECTED

        self.__connection_details = connection_details
        self.connection_id = connection_id

        if connection_details.client_id:
            try:
                self.ably.auth._configure_client_id(connection_details.client_id)
            except IncompatibleClientIdException as e:
                self.notify_state(ConnectionState.FAILED, reason=e)
                return

        if self.__state == ConnectionState.CONNECTED:
            state_change = ConnectionStateChange(ConnectionState.CONNECTED, ConnectionState.CONNECTED,
                                                 ConnectionEvent.UPDATE)
            self._emit(ConnectionEvent.UPDATE, state_change)
        else:
            self.notify_state(ConnectionState.CONNECTED, reason=reason)

        self.ably.channels._on_connected()

    async def on_disconnected(self, exception: AblyException) -> None:
        # RTN15h
        if self.transport:
            await self.transport.dispose()
        if exception:
            status_code = exception.status_code
            if status_code >= 500 and status_code <= 504:  # RTN17f1
                if len(self.__fallback_hosts) > 0:
                    try:
                        await self.connect_with_fallback_hosts(self.__fallback_hosts)
                    except Exception as e:
                        self.notify_state(self.__fail_state, reason=e)
                    return
                else:
                    log.info("No fallback host to try for disconnected protocol message")
            elif is_token_error(exception):
                await self.on_token_error(exception)
            else:
                self.notify_state(ConnectionState.DISCONNECTED, exception)
        else:
            log.warn("DISCONNECTED message received without error")

    async def on_token_error(self, exception: AblyException) -> None:
        if self.__error_reason is None or not is_token_error(self.__error_reason):
            self.__error_reason = exception
            try:
                await self.ably.auth._ensure_valid_auth_credentials(force=True)
            except Exception as e:
                self.on_error_from_authorize(e)
                return
            self.notify_state(self.__fail_state, exception, retry_immediately=True)
            return
        self.notify_state(self.__fail_state, exception)

    async def on_error(self, msg: dict, exception: AblyException) -> None:
        if msg.get("channel") is not None:  # RTN15i
            self.on_channel_message(msg)
            return
        if self.transport:
            await self.transport.dispose()
        if is_token_error(exception):  # RTN14b
            await self.on_token_error(exception)
        else:
            self.enact_state_change(ConnectionState.FAILED, exception)

    def on_error_from_authorize(self, exception: AblyException) -> None:
        log.info("ConnectionManager.on_error_from_authorize(): err = %s", exception)
        # RSA4a
        if exception.code == 40171:
            self.notify_state(ConnectionState.FAILED, exception)
        elif exception.status_code == 403:
            msg = 'Client configured authentication provider returned 403; failing the connection'
            log.error(f'ConnectionManager.on_error_from_authorize(): {msg}')
            self.notify_state(ConnectionState.FAILED, AblyException(msg, 403, 80019))
        else:
            msg = 'Client configured authentication provider request failed'
            log.warning(f'ConnectionManager.on_error_from_authorize: {msg}')
            self.notify_state(self.__fail_state, AblyException(msg, 401, 80019))

    async def on_closed(self) -> None:
        if self.transport:
            await self.transport.dispose()
        if self.connect_base_task:
            self.connect_base_task.cancel()

    def on_channel_message(self, msg: dict) -> None:
        self.__ably.channels._on_channel_message(msg)

    def on_heartbeat(self, id: Optional[str]) -> None:
        if self.__ping_future:
            # Resolve on heartbeat from ping request.
            if self.__ping_id == id:
                if not self.__ping_future.cancelled():
                    self.__ping_future.set_result(None)
                self.__ping_future = None

    def deactivate_transport(self, reason: Optional[AblyException] = None):
        self.transport = None
        self.notify_state(ConnectionState.DISCONNECTED, reason)

    def request_state(self, state: ConnectionState, force=False) -> None:
        log.debug(f'ConnectionManager.request_state(): state = {state}')

        if not force and state == self.state:
            return

        if state == ConnectionState.CONNECTING and self.__state == ConnectionState.CONNECTED:
            return

        if state == ConnectionState.CLOSING and self.__state == ConnectionState.CLOSED:
            return

        if state == ConnectionState.CONNECTING and self.__state in (ConnectionState.CLOSED,
                                                                    ConnectionState.FAILED):
            self.ably.channels._initialize_channels()

        if not force:
            self.enact_state_change(state)

        if state == ConnectionState.CONNECTING:
            self.start_connect()

        if state == ConnectionState.CLOSING:
            asyncio.create_task(self.close_impl())

    def start_connect(self) -> None:
        self.start_suspend_timer()
        self.start_transition_timer(ConnectionState.CONNECTING)
        self.connect_base_task = asyncio.create_task(self.connect_base())

    async def connect_with_fallback_hosts(self, fallback_hosts: list) -> Optional[Exception]:
        for host in fallback_hosts:
            try:
                if self.check_connection():
                    await self.try_host(host)
                    return
                else:
                    message = "Unable to connect, network unreachable"
                    log.exception(message)
                    exception = AblyException(message, status_code=404, code=80003)
                    self.notify_state(self.__fail_state, exception)
                    return
            except Exception as exc:
                exception = exc
                log.exception(f'Connection to {host} failed, reason={exception}')
        log.exception("No more fallback hosts to try")
        return exception

    async def connect_base(self) -> None:
        fallback_hosts = self.__fallback_hosts
        primary_host = self.options.get_realtime_host()
        try:
            await self.try_host(primary_host)
            return
        except Exception as exception:
            log.exception(f'Connection to {primary_host} failed, reason={exception}')
            if len(fallback_hosts) > 0:
                log.info("Attempting connection to fallback host(s)")
                resp = await self.connect_with_fallback_hosts(fallback_hosts)
                if not resp:
                    return
                exception = resp
            self.notify_state(self.__fail_state, reason=exception)

    async def try_host(self, host) -> None:
        try:
            params = await self.__get_transport_params()
        except AblyException as e:
            self.on_error_from_authorize(e)
            return
        self.transport = WebSocketTransport(self, host, params)
        self._emit('transport.pending', self.transport)
        self.transport.connect()

        future = asyncio.Future()

        def on_transport_connected():
            log.debug('ConnectionManager.try_a_host(): transport connected')
            if self.transport:
                self.transport.off('failed', on_transport_failed)
            if not future.done():
                future.set_result(None)

        async def on_transport_failed(exception):
            log.info('ConnectionManager.try_a_host(): transport failed')
            if self.transport:
                self.transport.off('connected', on_transport_connected)
                await self.transport.dispose()
            future.set_exception(exception)

        self.transport.once('connected', on_transport_connected)
        self.transport.once('failed', on_transport_failed)
        #  Fix asyncio CancelledError in python 3.7
        try:
            await future
        except asyncio.CancelledError:
            return

    def notify_state(self, state: ConnectionState, reason: Optional[AblyException] = None,
                     retry_immediately: Optional[bool] = None) -> None:
        # RTN15a
        retry_immediately = (retry_immediately is not False) and (
            state == ConnectionState.DISCONNECTED and self.__state == ConnectionState.CONNECTED)

        log.debug(
            f'ConnectionManager.notify_state(): new state: {state}'
            + ('; will retry immediately' if retry_immediately else '')
        )

        if state == self.__state:
            return

        self.cancel_transition_timer()
        self.check_suspend_timer(state)

        if retry_immediately:
            self.options.loop.call_soon(self.request_state, ConnectionState.CONNECTING)
        elif state == ConnectionState.DISCONNECTED:
            self.start_retry_timer(self.options.disconnected_retry_timeout)
        elif state == ConnectionState.SUSPENDED:
            self.start_retry_timer(self.options.suspended_retry_timeout)

        if (state == ConnectionState.DISCONNECTED and not retry_immediately) or state == ConnectionState.SUSPENDED:
            self.disconnect_transport()

        self.enact_state_change(state, reason)

        if state == ConnectionState.CONNECTED:
            self.send_queued_messages()
        elif state in (
            ConnectionState.CLOSING,
            ConnectionState.CLOSED,
            ConnectionState.SUSPENDED,
            ConnectionState.FAILED,
        ):
            self.fail_queued_messages(reason)
            self.ably.channels._propagate_connection_interruption(state, reason)

    def start_transition_timer(self, state: ConnectionState, fail_state: Optional[ConnectionState] = None) -> None:
        log.debug(f'ConnectionManager.start_transition_timer(): transition state = {state}')

        if self.transition_timer:
            log.debug('ConnectionManager.start_transition_timer(): clearing already-running timer')
            self.transition_timer.cancel()

        if fail_state is None:
            fail_state = self.__fail_state if state != ConnectionState.CLOSING else ConnectionState.CLOSED

        timeout = self.options.realtime_request_timeout

        def on_transition_timer_expire():
            if self.transition_timer:
                self.transition_timer = None
                log.info(f'ConnectionManager {state} timer expired, notifying new state: {fail_state}')
                self.notify_state(
                    fail_state,
                    AblyException("Connection cancelled due to request timeout", 504, 50003)
                )

        log.debug(f'ConnectionManager.start_transition_timer(): setting timer for {timeout}ms')

        self.transition_timer = Timer(timeout, on_transition_timer_expire)

    def cancel_transition_timer(self):
        log.debug('ConnectionManager.cancel_transition_timer()')
        if self.transition_timer:
            self.transition_timer.cancel()
            self.transition_timer = None

    def start_suspend_timer(self) -> None:
        log.debug('ConnectionManager.start_suspend_timer()')
        if self.suspend_timer:
            return

        def on_suspend_timer_expire() -> None:
            if self.suspend_timer:
                self.suspend_timer = None
                log.info('ConnectionManager suspend timer expired, requesting new state: suspended')
                self.notify_state(
                    ConnectionState.SUSPENDED,
                    AblyException("Connection to server unavailable", 400, 80002)
                )
                self.__fail_state = ConnectionState.SUSPENDED
                self.__connection_details = None

        self.suspend_timer = Timer(Defaults.connection_state_ttl, on_suspend_timer_expire)

    def check_suspend_timer(self, state: ConnectionState) -> None:
        if state not in (
            ConnectionState.CONNECTING,
            ConnectionState.DISCONNECTED,
            ConnectionState.SUSPENDED,
        ):
            self.cancel_suspend_timer()

    def cancel_suspend_timer(self) -> None:
        log.debug('ConnectionManager.cancel_suspend_timer()')
        self.__fail_state = ConnectionState.DISCONNECTED
        if self.suspend_timer:
            self.suspend_timer.cancel()
            self.suspend_timer = None

    def start_retry_timer(self, interval: int) -> None:
        def on_retry_timeout():
            log.info('ConnectionManager retry timer expired, retrying')
            self.retry_timer = None
            self.request_state(ConnectionState.CONNECTING)

        self.retry_timer = Timer(interval, on_retry_timeout)

    def cancel_retry_timer(self) -> None:
        if self.retry_timer:
            self.retry_timer.cancel()
            self.retry_timer = None

    def disconnect_transport(self) -> None:
        log.info('ConnectionManager.disconnect_transport()')
        if self.transport:
            self.disconnect_transport_task = asyncio.create_task(self.transport.dispose())

    async def on_auth_updated(self, token_details: TokenDetails):
        log.info(f"ConnectionManager.on_auth_updated(): state = {self.state}")
        if self.state == ConnectionState.CONNECTED:
            auth_message = {
                "action": ProtocolMessageAction.AUTH,
                "auth": {
                    "accessToken": token_details.token
                }
            }
            await self.send_protocol_message(auth_message)

            state_change = await self.once_async()

            if state_change.current == ConnectionState.CONNECTED:
                return
            elif state_change.current == ConnectionState.FAILED:
                raise state_change.reason
        elif self.state == ConnectionState.CONNECTING:
            if self.connect_base_task and not self.connect_base_task.done():
                self.connect_base_task.cancel()
            if self.transport:
                await self.transport.dispose()
        if self.state != ConnectionState.CONNECTED:
            future = asyncio.Future()

            def on_state_change(state_change: ConnectionStateChange) -> None:
                if state_change.current == ConnectionState.CONNECTED:
                    self.off('connectionstate', on_state_change)
                    future.set_result(token_details)
                if state_change.current in (
                        ConnectionState.CLOSED,
                        ConnectionState.FAILED,
                        ConnectionState.SUSPENDED
                ):
                    self.off('connectionstate', on_state_change)
                    future.set_exception(state_change.reason or self.get_state_error())

            self.on('connectionstate', on_state_change)

            if self.state == ConnectionState.CONNECTING:
                self.start_connect()
            else:
                self.request_state(ConnectionState.CONNECTING)

            return await future

    @property
    def ably(self):
        return self.__ably

    @property
    def state(self) -> ConnectionState:
        return self.__state

    @property
    def connection_details(self) -> Optional[ConnectionDetails]:
        return self.__connection_details




class InstallPycrypto:
    def __getattr__(self, name):
        raise ImportError(
            "This requires to install ably with crypto support: pip install 'ably[crypto]'"
        )


AES = Random = InstallPycrypto()



import inspect
import random
import string
import asyncio
import time
from typing import Callable


def get_random_id():
    # get random string of letters and digits
    source = string.ascii_letters + string.digits
    random_id = ''.join((random.choice(source) for i in range(8)))
    return random_id


def is_callable_or_coroutine(value):
    return asyncio.iscoroutinefunction(value) or inspect.isfunction(value) or inspect.ismethod(value)


def unix_time_ms():
    return round(time.time_ns() / 1_000_000)


def is_token_error(exception):
    return 40140 <= exception.code < 40150


class Timer:
    def __init__(self, timeout: float, callback: Callable):
        self._timeout = timeout
        self._callback = callback
        self._task = asyncio.create_task(self._job())

    async def _job(self):
        await asyncio.sleep(self._timeout / 1000)
        if asyncio.iscoroutinefunction(self._callback):
            await self._callback()
        else:
            self._callback()

    def cancel(self):
        self._task.cancel()



from dataclasses import dataclass
from typing import Optional
from enum import Enum
from ably.util.exceptions import AblyException


class ChannelState(str, Enum):
    INITIALIZED = 'initialized'
    ATTACHING = 'attaching'
    ATTACHED = 'attached'
    DETACHING = 'detaching'
    DETACHED = 'detached'
    SUSPENDED = 'suspended'
    FAILED = 'failed'


@dataclass
class ChannelStateChange:
    previous: ChannelState
    current: ChannelState
    resumed: bool
    reason: Optional[AblyException] = None



import logging
import asyncio
from typing import Optional
from ably.sync.realtime.realtime_channel import ChannelsSync
from ably.sync.realtime.connection import Connection, ConnectionState
from ably.sync.rest.rest import AblyRestSync


log = logging.getLogger(__name__)


class AblyRealtime(AblyRestSync):
    """
    Ably Realtime Client

    Attributes
    ----------
    loop: AbstractEventLoop
        asyncio running event loop
    auth: Auth
        authentication object
    options: Options
        auth options object
    connection: Connection
        realtime connection object
    channels: Channels
        realtime channel object

    Methods
    -------
    connect()
        Establishes the realtime connection
    close()
        Closes the realtime connection
    """

    def __init__(self, key: Optional[str] = None, loop: Optional[asyncio.AbstractEventLoop] = None, **kwargs):
        """Constructs a RealtimeClient object using an Ably API key.

        Parameters
        ----------
        key: str
            A valid ably API key string
        loop: AbstractEventLoop, optional
            asyncio running event loop
        auto_connect: bool
            When true, the client connects to Ably as soon as it is instantiated.
            You can set this to false and explicitly connect to Ably using the
            connect() method. The default is true.
        **kwargs: client options
            realtime_host: str
                Enables a non-default Ably host to be specified for realtime connections.
                For development environments only. The default value is realtime.ably.io.
            environment: str
                Enables a custom environment to be used with the Ably service. Defaults to `production`
            realtime_request_timeout: float
                Timeout (in milliseconds) for the wait of acknowledgement for operations performed via a realtime
                connection. Operations include establishing a connection with Ably, or sending a HEARTBEAT,
                CONNECT, ATTACH, DETACH or CLOSE request. The default is 10 seconds(10000 milliseconds).
            disconnected_retry_timeout: float
                If the connection is still in the DISCONNECTED state after this delay, the client library will
                attempt to reconnect automatically. The default is 15 seconds.
            channel_retry_timeout: float
                When a channel becomes SUSPENDED following a server initiated DETACHED, after this delay, if the
                channel is still SUSPENDED and the connection is in CONNECTED, the client library will attempt to
                re-attach the channel automatically. The default is 15 seconds.
            fallback_hosts: list[str]
                An array of fallback hosts to be used in the case of an error necessitating the use of an
                alternative host. If you have been provided a set of custom fallback hosts by Ably, please specify
                them here.
            connection_state_ttl: float
                The duration that Ably will persist the connection state for when a Realtime client is abruptly
                disconnected.
            suspended_retry_timeout: float
                When the connection enters the SUSPENDED state, after this delay, if the state is still SUSPENDED,
                the client library attempts to reconnect automatically. The default is 30 seconds.
            connectivity_check_url: string
                Override the URL used by the realtime client to check if the internet is available.
                In the event of a failure to connect to the primary endpoint, the client will send a
                GET request to this URL to check if the internet is available. If this request returns
                a success response the client will attempt to connect to a fallback host.
        Raises
        ------
        ValueError
            If no authentication key is not provided
        """

        if loop is None:
            try:
                loop = asyncio.get_running_loop()
            except RuntimeError:
                log.warning('Realtime client created outside event loop')

        self._is_realtime: bool = True

        # RTC1
        super().__init__(key, loop=loop, **kwargs)

        self.key = key
        self.__connection = Connection(self)
        self.__channels = ChannelsSync(self)

        # RTN3
        if self.options.auto_connect:
            self.connection.connection_manager.request_state(ConnectionState.CONNECTING, force=True)

    # RTC15
    def connect(self) -> None:
        """Establishes a realtime connection.

        Explicitly calling connect() is unnecessary unless the autoConnect attribute of the ClientOptions object
        is false. Unless already connected or connecting, this method causes the connection to open, entering the
        CONNECTING state.
        """
        log.info('Realtime.connect() called')
        # RTC15a
        self.connection.connect()

    # RTC16
    def close(self) -> None:
        """Causes the connection to close, entering the closing state.
        Once closed, the library will not attempt to re-establish the
        connection without an explicit call to connect()
        """
        log.info('Realtime.close() called')
        # RTC16a
        self.connection.close()
        super().close()

    # RTC2
    @property
    def connection(self) -> Connection:
        """Returns the realtime connection object"""
        return self.__connection

    # RTC3, RTS1
    @property
    def channels(self) -> ChannelsSync:
        """Returns the realtime channel object"""
        return self.__channels



from __future__ import annotations
import functools
import logging
from ably.realtime.connectionmanager import ConnectionManager
from ably.types.connectiondetails import ConnectionDetails
from ably.types.connectionstate import ConnectionEvent, ConnectionState, ConnectionStateChange
from ably.util.eventemitter import EventEmitter
from ably.util.exceptions import AblyException
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from ably.realtime.realtime import AblyRealtime

log = logging.getLogger(__name__)


class Connection(EventEmitter):  # RTN4
    """Ably Realtime Connection

    Enables the management of a connection to Ably

    Attributes
    ----------
    state: str
        Connection state
    error_reason: ErrorInfo
        An ErrorInfo object describing the last error which occurred on the channel, if any.


    Methods
    -------
    connect()
        Establishes a realtime connection
    close()
        Closes a realtime connection
    ping()
        Pings a realtime connection
    """

    def __init__(self, realtime: AblyRealtime):
        self.__realtime = realtime
        self.__error_reason: Optional[AblyException] = None
        self.__state = ConnectionState.CONNECTING if realtime.options.auto_connect else ConnectionState.INITIALIZED
        self.__connection_manager = ConnectionManager(self.__realtime, self.state)
        self.__connection_manager.on('connectionstate', self._on_state_update)  # RTN4a
        self.__connection_manager.on('update', self._on_connection_update)  # RTN4h
        super().__init__()

    # RTN11
    def connect(self) -> None:
        """Establishes a realtime connection.

        Causes the connection to open, entering the connecting state
        """
        self.__error_reason = None
        self.connection_manager.request_state(ConnectionState.CONNECTING)

    async def close(self) -> None:
        """Causes the connection to close, entering the closing state.

        Once closed, the library will not attempt to re-establish the
        connection without an explicit call to connect()
        """
        self.connection_manager.request_state(ConnectionState.CLOSING)
        await self.once_async(ConnectionState.CLOSED)

    # RTN13
    async def ping(self) -> float:
        """Send a ping to the realtime connection

        When connected, sends a heartbeat ping to the Ably server and executes
        the callback with any error and the response time in milliseconds when
        a heartbeat ping request is echoed from the server.

        Raises
        ------
        AblyException
            If ping request cannot be sent due to invalid state

        Returns
        -------
        float
            The response time in milliseconds
        """
        return await self.__connection_manager.ping()

    def _on_state_update(self, state_change: ConnectionStateChange) -> None:
        log.info(f'Connection state changing from {self.state} to {state_change.current}')
        self.__state = state_change.current
        if state_change.reason is not None:
            self.__error_reason = state_change.reason
        self.__realtime.options.loop.call_soon(functools.partial(self._emit, state_change.current, state_change))

    def _on_connection_update(self, state_change: ConnectionStateChange) -> None:
        self.__realtime.options.loop.call_soon(functools.partial(self._emit, ConnectionEvent.UPDATE, state_change))

    # RTN4d
    @property
    def state(self) -> ConnectionState:
        """The current connection state of the connection"""
        return self.__state

    # RTN25
    @property
    def error_reason(self) -> Optional[AblyException]:
        """An object describing the last error which occurred on the channel, if any."""
        return self.__error_reason

    @state.setter
    def state(self, value: ConnectionState) -> None:
        self.__state = value

    @property
    def connection_manager(self) -> ConnectionManager:
        return self.__connection_manager

    @property
    def connection_details(self) -> Optional[ConnectionDetails]:
        return self.__connection_manager.connection_details



import base64
import json
import logging

from ably.sync.types.typedbuffer import TypedBuffer
from ably.sync.types.mixins import EncodeDataMixin
from ably.sync.util.crypto import CipherData
from ably.sync.util.exceptions import AblyException

log = logging.getLogger(__name__)


def to_text(value):
    if value is None:
        return value
    elif isinstance(value, str):
        return value
    elif isinstance(value, bytes):
        return value.decode()
    else:
        raise TypeError("expected string or bytes, not %s" % type(value))


class Message(EncodeDataMixin):

    def __init__(self,
                 name=None,  # TM2g
                 data=None,  # TM2d
                 client_id=None,  # TM2b
                 id=None,  # TM2a
                 connection_id=None,  # TM2c
                 connection_key=None,  # TM2h
                 encoding='',  # TM2e
                 timestamp=None,  # TM2f
                 extras=None,  # TM2i
                 ):

        super().__init__(encoding)

        self.__name = to_text(name)
        self.__data = data
        self.__client_id = to_text(client_id)
        self.__id = to_text(id)
        self.__connection_id = connection_id
        self.__connection_key = connection_key
        self.__timestamp = timestamp
        self.__extras = extras

    def __eq__(self, other):
        if isinstance(other, Message):
            return (self.name == other.name
                    and self.data == other.data
                    and self.client_id == other.client_id
                    and self.timestamp == other.timestamp)
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, Message):
            result = self.__eq__(other)
            if result != NotImplemented:
                return not result
        return NotImplemented

    @property
    def name(self):
        return self.__name

    @property
    def data(self):
        return self.__data

    @property
    def client_id(self):
        return self.__client_id

    @property
    def id(self):
        return self.__id

    @id.setter
    def id(self, value):
        self.__id = value

    @property
    def connection_id(self):
        return self.__connection_id

    @property
    def connection_key(self):
        return self.__connection_key

    @property
    def timestamp(self):
        return self.__timestamp

    @property
    def extras(self):
        return self.__extras

    def encrypt(self, channel_cipher):
        if isinstance(self.data, CipherData):
            return

        elif isinstance(self.data, str):
            self._encoding_array.append('utf-8')

        if isinstance(self.data, dict) or isinstance(self.data, list):
            self._encoding_array.append('json')
            self._encoding_array.append('utf-8')

        typed_data = TypedBuffer.from_obj(self.data)
        if typed_data.buffer is None:
            return True
        encrypted_data = channel_cipher.encrypt(typed_data.buffer)
        self.__data = CipherData(encrypted_data, typed_data.type,
                                 cipher_type=channel_cipher.cipher_type)

    @staticmethod
    def decrypt_data(channel_cipher, data):
        if not isinstance(data, CipherData):
            return
        decrypted_data = channel_cipher.decrypt(data.buffer)
        decrypted_typed_buffer = TypedBuffer(decrypted_data, data.type)

        return decrypted_typed_buffer.decode()

    def decrypt(self, channel_cipher):
        decrypted_data = self.decrypt_data(channel_cipher, self.__data)
        if decrypted_data is not None:
            self.__data = decrypted_data

    def as_dict(self, binary=False):
        data = self.data
        data_type = None
        encoding = self._encoding_array[:]

        if isinstance(data, (dict, list)):
            encoding.append('json')
            data = json.dumps(data)
            data = str(data)
        elif isinstance(data, str) and not binary:
            pass
        elif not binary and isinstance(data, (bytearray, bytes)):
            data = base64.b64encode(data).decode('ascii')
            encoding.append('base64')
        elif isinstance(data, CipherData):
            encoding.append(data.encoding_str)
            data_type = data.type
            if not binary:
                data = base64.b64encode(data.buffer).decode('ascii')
                encoding.append('base64')
            else:
                data = data.buffer
        elif binary and isinstance(data, bytearray):
            data = bytes(data)

        if not (isinstance(data, (bytes, str, list, dict, bytearray)) or data is None):
            raise AblyException("Invalid data payload", 400, 40011)

        request_body = {
            'name': self.name,
            'data': data,
            'timestamp': self.timestamp or None,
            'type': data_type or None,
            'clientId': self.client_id or None,
            'id': self.id or None,
            'connectionId': self.connection_id or None,
            'connectionKey': self.connection_key or None,
            'extras': self.extras,
        }

        if encoding:
            request_body['encoding'] = '/'.join(encoding).strip('/')

        # None values aren't included
        request_body = {k: v for k, v in request_body.items() if v is not None}

        return request_body

    @staticmethod
    def from_encoded(obj, cipher=None):
        id = obj.get('id')
        name = obj.get('name')
        data = obj.get('data')
        client_id = obj.get('clientId')
        connection_id = obj.get('connectionId')
        timestamp = obj.get('timestamp')
        encoding = obj.get('encoding', '')
        extras = obj.get('extras', None)

        decoded_data = Message.decode(data, encoding, cipher)

        return Message(
            id=id,
            name=name,
            connection_id=connection_id,
            client_id=client_id,
            timestamp=timestamp,
            extras=extras,
            **decoded_data
        )

    @staticmethod
    def __update_empty_fields(proto_msg: dict, msg: dict, msg_index: int):
        if msg.get("id") is None or msg.get("id") == '':
            msg['id'] = f"{proto_msg.get('id')}:{msg_index}"
        if msg.get("connectionId") is None or msg.get("connectionId") == '':
            msg['connectionId'] = proto_msg.get('connectionId')
        if msg.get("timestamp") is None or msg.get("timestamp") == 0:
            msg['timestamp'] = proto_msg.get('timestamp')

    @staticmethod
    def update_inner_message_fields(proto_msg: dict):
        messages: list[dict] = proto_msg.get('messages')
        presence_messages: list[dict] = proto_msg.get('presence')
        if messages is not None:
            msg_index = 0
            for msg in messages:
                Message.__update_empty_fields(proto_msg, msg, msg_index)
                msg_index = msg_index + 1

        if presence_messages is not None:
            msg_index = 0
            for presence_msg in presence_messages:
                Message.__update_empty_fields(proto_msg, presence_msg, msg_index)
                msg_index = msg_index + 1


def make_message_response_handler(cipher):
    def encrypted_message_response_handler(response):
        messages = response.to_native()
        return Message.from_encoded_array(messages, cipher=cipher)
    return encrypted_message_response_handler



import base64
from collections import OrderedDict
import logging
import json
import os
from typing import Iterator
from urllib import parse

from methoddispatch import SingleDispatch, singledispatch
import msgpack

from ably.http.paginatedresult import PaginatedResult, format_params
from ably.types.channeldetails import ChannelDetails
from ably.types.message import Message, make_message_response_handler
from ably.types.presence import Presence
from ably.util.crypto import get_cipher
from ably.util.exceptions import catch_all, IncompatibleClientIdException

log = logging.getLogger(__name__)


class Channel(SingleDispatch):
    def __init__(self, ably, name, options):
        self.__ably = ably
        self.__name = name
        self.__base_path = '/channels/%s/' % parse.quote_plus(name, safe=':')
        self.__cipher = None
        self.options = options
        self.__presence = Presence(self)

    @catch_all
    async def history(self, direction=None, limit: int = None, start=None, end=None):
        """Returns the history for this channel"""
        params = format_params({}, direction=direction, start=start, end=end, limit=limit)
        path = self.__base_path + 'messages' + params

        message_handler = make_message_response_handler(self.__cipher)
        return await PaginatedResult.paginated_query(
            self.ably.http, url=path, response_processor=message_handler)

    def __publish_request_body(self, messages):
        """
        Helper private method, separated from publish() to test RSL1j
        """
        # Idempotent publishing
        if self.ably.options.idempotent_rest_publishing:
            # RSL1k1
            if all(message.id is None for message in messages):
                base_id = base64.b64encode(os.urandom(12)).decode()
                for serial, message in enumerate(messages):
                    message.id = '{}:{}'.format(base_id, serial)

        request_body_list = []
        for m in messages:
            if m.client_id == '*':
                raise IncompatibleClientIdException(
                    'Wildcard client_id is reserved and cannot be used when publishing messages',
                    400, 40012)
            elif m.client_id is not None and not self.ably.auth.can_assume_client_id(m.client_id):
                raise IncompatibleClientIdException(
                    'Cannot publish with client_id \'{}\' as it is incompatible with the '
                    'current configured client_id \'{}\''.format(m.client_id, self.ably.auth.client_id),
                    400, 40012)

            if self.cipher:
                m.encrypt(self.__cipher)

            request_body_list.append(m)

        request_body = [
            message.as_dict(binary=self.ably.options.use_binary_protocol)
            for message in request_body_list]

        if len(request_body) == 1:
            request_body = request_body[0]

        return request_body

    @singledispatch
    def _publish(self, arg, *args, **kwargs):
        raise TypeError('Unexpected type %s' % type(arg))

    @_publish.register(Message)
    async def publish_message(self, message, params=None, timeout=None):
        return await self.publish_messages([message], params, timeout=timeout)

    @_publish.register(list)
    async def publish_messages(self, messages, params=None, timeout=None):
        request_body = self.__publish_request_body(messages)
        if not self.ably.options.use_binary_protocol:
            request_body = json.dumps(request_body, separators=(',', ':'))
        else:
            request_body = msgpack.packb(request_body, use_bin_type=True)

        path = self.__base_path + 'messages'
        if params:
            params = {k: str(v).lower() if type(v) is bool else v for k, v in params.items()}
            path += '?' + parse.urlencode(params)
        return await self.ably.http.post(path, body=request_body, timeout=timeout)

    @_publish.register(str)
    async def publish_name_data(self, name, data, timeout=None):
        messages = [Message(name, data)]
        return await self.publish_messages(messages, timeout=timeout)

    async def publish(self, *args, **kwargs):
        """Publishes a message on this channel.

        :Parameters:
        - `name`: the name for this message.
        - `data`: the data for this message.
        - `messages`: list of `Message` objects to be published.
        - `message`: a single `Message` objet to be published

        :attention: You can publish using `name` and `data` OR `messages` OR
        `message`, never all three.
        """
        # For backwards compatibility
        if len(args) == 0:
            if len(kwargs) == 0:
                return await self.publish_name_data(None, None)

            if 'name' in kwargs or 'data' in kwargs:
                name = kwargs.pop('name', None)
                data = kwargs.pop('data', None)
                return await self.publish_name_data(name, data, **kwargs)

            if 'messages' in kwargs:
                messages = kwargs.pop('messages')
                return await self.publish_messages(messages, **kwargs)

        return await self._publish(*args, **kwargs)

    async def status(self):
        """Retrieves current channel active status with no. of publishers, subscribers, presence_members etc"""

        path = '/channels/%s' % self.name
        response = await self.ably.http.get(path)
        obj = response.to_native()
        return ChannelDetails.from_dict(obj)

    @property
    def ably(self):
        return self.__ably

    @property
    def name(self):
        return self.__name

    @property
    def base_path(self):
        return self.__base_path

    @property
    def cipher(self):
        return self.__cipher

    @property
    def options(self):
        return self.__options

    @property
    def presence(self):
        return self.__presence

    @options.setter
    def options(self, options):
        self.__options = options

        if options and 'cipher' in options:
            cipher = options.get('cipher')
            if cipher is not None:
                cipher = get_cipher(cipher)
            self.__cipher = cipher


class Channels:
    def __init__(self, rest):
        self.__ably = rest
        self.__all: dict = OrderedDict()

    def get(self, name, **kwargs):
        if isinstance(name, bytes):
            name = name.decode('ascii')

        if name not in self.__all:
            result = self.__all[name] = Channel(self.__ably, name, kwargs)
        else:
            result = self.__all[name]
            if len(kwargs) != 0:
                result.options = kwargs

        return result

    def __getitem__(self, key):
        return self.get(key)

    def __getattr__(self, name):
        return self.get(name)

    def __contains__(self, item):
        if isinstance(item, Channel):
            name = item.name
        elif isinstance(item, bytes):
            name = item.decode('ascii')
        else:
            name = item

        return name in self.__all

    def __iter__(self) -> Iterator[str]:
        return iter(self.__all.values())

    # RSN4
    def release(self, name: str):
        """Releases a Channel object, deleting it, and enabling it to be garbage collected.
        If the channel does not exist, nothing happens.

        It also removes any listeners associated with the channel.

        Parameters
        ----------
        name: str
            Channel name
        """

        if name not in self.__all:
            return
        del self.__all[name]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ably',
 'ably.http',
 'ably.realtime',
 'ably.rest',
 'ably.scripts',
 'ably.sync',
 'ably.sync.http',
 'ably.sync.realtime',
 'ably.sync.rest',
 'ably.sync.scripts',
 'ably.sync.transport',
 'ably.sync.types',
 'ably.sync.util',
 'ably.transport',
 'ably.types',
 'ably.util']

package_data = \
{'': ['*']}

install_requires = \
['httpx[http2]>=0.24.1,<0.25.0',
 'methoddispatch>=3.0.2,<4.0.0',
 'msgpack>=1.0.0,<2.0.0',
 'pyee>=9.0.4,<10.0.0',
 'websockets>=10.3,<11.0']

extras_require = \
{'crypto': ['pycryptodome'], 'oldcrypto': ['pycrypto>=2.6.1,<3.0.0']}

entry_points = \
{'console_scripts': ['unasync = ably.scripts.unasync:run']}

setup_kwargs = {
    'name': 'ably',
    'version': '2.0.2',
    'description': 'Python REST and Realtime client library SDK for Ably realtime messaging service',
    'long_description': 'Official Ably Bindings for Python\n==================================\n\nA Python client library for Ably Realtime messaging.\n\n\nSetup\n-----\n\nYou can install this package by using the pip tool and installing:\n\n    pip install ably\n\n\nUsing Ably for Python\n---------------------\n\n- Sign up for Ably at https://ably.com/sign-up\n- Get usage examples at https://github.com/ably/ably-python\n- Visit https://ably.com/docs for a complete API reference and more examples\n',
    'author': 'Ably',
    'author_email': 'support@ably.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://ably.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from enum import Enum


class Flag(int, Enum):
    # Channel attach state flags
    HAS_PRESENCE = 1 << 0
    HAS_BACKLOG = 1 << 1
    RESUMED = 1 << 2
    TRANSIENT = 1 << 4
    ATTACH_RESUME = 1 << 5
    # Channel mode flags
    PRESENCE = 1 << 16
    PUBLISH = 1 << 17
    SUBSCRIBE = 1 << 18
    PRESENCE_SUBSCRIBE = 1 << 19


def has_flag(message_flags: int, flag: Flag):
    return message_flags & flag > 0



from datetime import datetime, timedelta
from urllib import parse

from ably.http.paginatedresult import PaginatedResult
from ably.types.mixins import EncodeDataMixin


def _ms_since_epoch(dt):
    epoch = datetime.utcfromtimestamp(0)
    delta = dt - epoch
    return int(delta.total_seconds() * 1000)


def _dt_from_ms_epoch(ms):
    epoch = datetime.utcfromtimestamp(0)
    return epoch + timedelta(milliseconds=ms)


class PresenceAction:
    ABSENT = 0
    PRESENT = 1
    ENTER = 2
    LEAVE = 3
    UPDATE = 4


class PresenceMessage(EncodeDataMixin):

    def __init__(self,
                 id=None,  # TP3a
                 action=None,  # TP3b
                 client_id=None,  # TP3c
                 connection_id=None,  # TP3d
                 data=None,  # TP3e
                 encoding=None,  # TP3f
                 timestamp=None,  # TP3g
                 member_key=None,  # TP3h (for RT only)
                 extras=None,  # TP3i (functionality not specified)
                 ):

        self.__id = id
        self.__action = action
        self.__client_id = client_id
        self.__connection_id = connection_id
        self.__data = data
        self.__encoding = encoding
        self.__timestamp = timestamp
        self.__member_key = member_key
        self.__extras = extras

    @property
    def id(self):
        return self.__id

    @property
    def action(self):
        return self.__action

    @property
    def client_id(self):
        return self.__client_id

    @property
    def connection_id(self):
        return self.__connection_id

    @property
    def data(self):
        return self.__data

    @property
    def encoding(self):
        return self.__encoding

    @property
    def timestamp(self):
        return self.__timestamp

    @property
    def member_key(self):
        if self.connection_id and self.client_id:
            return "%s:%s" % (self.connection_id, self.client_id)

    @property
    def extras(self):
        return self.__extras

    @staticmethod
    def from_encoded(obj, cipher=None):
        id = obj.get('id')
        action = obj.get('action', PresenceAction.ENTER)
        client_id = obj.get('clientId')
        connection_id = obj.get('connectionId')
        data = obj.get('data')
        encoding = obj.get('encoding', '')
        timestamp = obj.get('timestamp')
        # member_key = obj.get('memberKey', None)
        extras = obj.get('extras', None)

        if timestamp is not None:
            timestamp = _dt_from_ms_epoch(timestamp)

        decoded_data = PresenceMessage.decode(data, encoding, cipher)

        return PresenceMessage(
            id=id,
            action=action,
            client_id=client_id,
            connection_id=connection_id,
            timestamp=timestamp,
            extras=extras,
            **decoded_data
        )


class Presence:
    def __init__(self, channel):
        self.__base_path = '/channels/%s/' % parse.quote_plus(channel.name)
        self.__binary = channel.ably.options.use_binary_protocol
        self.__http = channel.ably.http
        self.__cipher = channel.cipher

    def _path_with_qs(self, rel_path, qs=None):
        path = rel_path
        if qs:
            path += ('?' + parse.urlencode(qs))
        return path

    async def get(self, limit=None):
        qs = {}
        if limit:
            if limit > 1000:
                raise ValueError("The maximum allowed limit is 1000")
            qs['limit'] = limit
        path = self._path_with_qs(self.__base_path + 'presence', qs)

        presence_handler = make_presence_response_handler(self.__cipher)
        return await PaginatedResult.paginated_query(
            self.__http, url=path, response_processor=presence_handler)

    async def history(self, limit=None, direction=None, start=None, end=None):
        qs = {}
        if limit:
            if limit > 1000:
                raise ValueError("The maximum allowed limit is 1000")
            qs['limit'] = limit
        if direction:
            qs['direction'] = direction
        if start:
            if isinstance(start, int):
                qs['start'] = start
            else:
                qs['start'] = _ms_since_epoch(start)
        if end:
            if isinstance(end, int):
                qs['end'] = end
            else:
                qs['end'] = _ms_since_epoch(end)

        if 'start' in qs and 'end' in qs and qs['start'] > qs['end']:
            raise ValueError("'end' parameter has to be greater than or equal to 'start'")

        path = self._path_with_qs(self.__base_path + 'presence/history', qs)

        presence_handler = make_presence_response_handler(self.__cipher)
        return await PaginatedResult.paginated_query(
            self.__http, url=path, response_processor=presence_handler)


def make_presence_response_handler(cipher):
    def encrypted_presence_response_handler(response):
        messages = response.to_native()
        return PresenceMessage.from_encoded_array(messages, cipher=cipher)
    return encrypted_presence_response_handler
